
class ElasticsearchRavenError(Exception):
    pass


class DamagedSentryMessageError(ElasticsearchRavenError):
    pass


class BadSentryMessageHeaderError(ElasticsearchRavenError):
    pass


class DamagedSentryMessageBodyError(DamagedSentryMessageError):
    pass



import base64

try:
    import queue
except ImportError:
    import Queue as queue

import kombu

from elasticsearch_raven import transport


class Empty(Exception):
    pass


class AbstractQueue:
    def get(self, timeout=None):
        raise NotImplementedError

    def put(self, message):
        raise NotImplementedError

    def join(self):
        raise NotImplementedError

    def task_done(self):
        raise NotImplementedError

    def has_nonpersistent_task(self):
        raise NotImplementedError


class ThreadingQueue:
    def __init__(self, *args, **kwargs):
        self.queue = queue.Queue(*args, **kwargs)

    def get(self, timeout=None):
        try:
            return self.queue.get(timeout=timeout)
        except queue.Empty:
            raise Empty()

    def put(self, message):
        self.queue.put(message)

    def join(self):
        self.queue.join()

    def task_done(self):
        self.queue.task_done()

    def has_nonpersistent_task(self):
        return bool(self.queue.unfinished_tasks)


class KombuQueue:
    def __init__(self, amqp_url, queue_name):
        self.connection = kombu.Connection(amqp_url)
        self.queue = self.connection.SimpleQueue(queue_name)
        self._processed = None

    def get(self, timeout=None):
        try:
            self._processed = self.queue.get(timeout=timeout)
        except self.queue.Empty:
            raise Empty()
        else:
            return self._deserialize(self._processed.payload)

    def put(self, message):
        return self.queue.put(self._serialize(message))

    def join(self):
        pass

    def task_done(self):
        self._processed.ack()
        self._processed = None

    def has_nonpersistent_task(self):
        return False

    def _serialize(self, message):
        return message.headers, base64.b64encode(message.body).decode('utf-8')

    def _deserialize(self, data):
        headers, encoded_body = data
        body = base64.b64decode(encoded_body.encode('utf-8'))
        return transport.SentryMessage(headers, body)



try:
    import queue
except ImportError:
    import Queue as queue

from elasticsearch_raven import configuration
from elasticsearch_raven import transport
from elasticsearch_raven.udp_server import Sender


class HttpUtils:
    def __init__(self):
        self._pending_logs = queue.Queue(configuration['queue_maxsize'])
        self._exception_queue = queue.Queue()

    def start_sender(self):
        log_transport = transport.get_configured_log_transport()
        sender = Sender(log_transport, self._pending_logs,
                        self._exception_queue).as_thread()
        sender.start()

    def get_application(self):
        def application(environ, start_response):
            try:
                exception = self._exception_queue.get_nowait()
            except queue.Empty:
                pass
            else:
                raise exception
            length = int(environ.get('CONTENT_LENGTH', '0'))
            data = environ['wsgi.input'].read(length)
            self._pending_logs.put(transport.SentryMessage.create_from_http(
                environ['HTTP_X_SENTRY_AUTH'], data))

            status = '200 OK'
            response_headers = [('Content-Type', 'text/plain')]
            start_response(status, response_headers)
            return [''.encode('utf-8')]
        return application



import itertools
import collections


def postfix_encoded_data(encoded_data):
    field_names_to_postfix = ['extra']
    sentry_fields = keys_starting_with(encoded_data, 'sentry.')
    field_names_to_postfix.extend(sentry_fields)

    fields_to_postfix = filter(lambda x: x in field_names_to_postfix,
                               encoded_data)

    for field in fields_to_postfix:
        _, encoded_data[field] = next(postfix_types(
            ('', encoded_data[field])))


def keys_starting_with(dictionary, word):
    return (key for key in dictionary.keys() if key.startswith(word))


def postfix_types(row):
    name, data = row
    if data is None:
        return postfix_none(name, data)
    type_postfix = postfixes.get(type(data), postfix_other)
    return type_postfix(name, data)


def postfix_none(name, _):
    yield name, None


def postfix_other(name, data):
    yield ('%s<%s>' % (name, type(data).__name__)), data


def postfix_dict(name, data):
    if name.endswith(">"):
        name += "<dict>"
    postfix_items = list(map(postfix_types, data.items()))
    yield name, dict(itertools.chain(*postfix_items))


def postfix_str(name, data):
    yield name + '<string>', data


def postfix_list(name, data):
    for k, v in _split_list_by_type(data).items():
        yield name + k, v


def _split_list_by_type(data):
    result = collections.defaultdict(list)
    for element in data:
        for value_type, value in postfix_types(('', element)):
            result[value_type].append(value)
    if result:
        return result
    else:
        return {'': []}


postfixes = {
    dict: postfix_dict,
    str: postfix_str,
    list: postfix_list,
}



import os

configuration = {
    'host': os.environ.get('ELASTICSEARCH_HOST', 'localhost:9200'),
    'use_ssl': os.environ.get('USE_SSL', False),
    'queue_maxsize': os.environ.get('QUEUE_MAXSIZE', 1000),
    'http_auth': os.environ.get('ELASTICSEARCH_AUTH', None),
    'amqp_url': os.environ.get('AMQP_URL',
                               'amqp://guest:guest@localhost:5672//'),
    'amqp_queue': os.environ.get('AMQP_QUEUE', 'elasticsearch-raven-logs'),
}



import argparse
import datetime
import socket
import sys
import time
import threading

try:
    import queue
except ImportError:
    import Queue as queue

import elasticsearch

from elasticsearch_raven import configuration
from elasticsearch_raven import transport
from elasticsearch_raven import queues


def run_server():
    args = _parse_args()
    try:
        sock = get_socket(args.ip, args.port)
    except socket.gaierror:
        sys.stdout.write('Wrong hostname.\n')
        sys.exit(1)
    else:
        log_transport = transport.get_configured_log_transport()
        if args.amqp_queue:
            pending_logs = queues.KombuQueue(configuration['amqp_url'],
                                             configuration['amqp_queue'])
        else:
            pending_logs = queues.ThreadingQueue(
                configuration['queue_maxsize'])
        Server(sock, pending_logs, log_transport,
               args.debug).run()


def _parse_args():
    parser = argparse.ArgumentParser(description='Udp proxy server for raven')
    parser.add_argument('ip')
    parser.add_argument('port', type=int)
    parser.add_argument('--debug', dest='debug', action='store_const',
                        const=True, default=False,
                        help='Print debug logs to stdout')
    parser.add_argument(
        '--amqp-queue', action='store_const', const=True, default=False,
        help='Use amqp queue to store logs to send to elasticsearch.')
    return parser.parse_args()


def get_socket(ip, port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.bind((ip, int(port)))
    return sock


class Server(object):
    def __init__(self, sock, pending_logs, log_transport, debug=False):
        self.sock = sock
        self.pending_logs = pending_logs
        self.log_transport = log_transport
        self.debug = debug
        self.exception_queue = queue.Queue()

    def run(self):
        handler = Handler(self.sock, self.pending_logs, self.exception_queue,
                          debug=self.debug).as_thread()
        sender = Sender(self.log_transport, self.pending_logs,
                        self.exception_queue).as_thread()
        handler.start()
        sender.start()
        try:
            raise self.exception_queue.get()
        except KeyboardInterrupt:
            self.sock.close()
            try:
                while self.pending_logs.has_nonpersistent_task():
                    try:
                        raise self.exception_queue.get(timeout=1)
                    except queue.Empty:
                        pass
            except KeyboardInterrupt:
                pass


class Handler(object):
    def __init__(self, sock, pending_logs, exception_queue, debug=False):
        self.sock = sock
        self.pending_logs = pending_logs
        self.exception_queue = exception_queue
        self.debug = debug

    def as_thread(self):
        handler = threading.Thread(target=self._handle)
        handler.daemon = True
        return handler

    def _handle(self):
        try:
            while True:
                data, address = self.sock.recvfrom(65535)
                message = transport.SentryMessage.create_from_udp(data)
                self.pending_logs.put(message)
                if self.debug:
                    sys.stdout.write('{host}:{port} [{date}]\n'.format(
                        host=address[0], port=address[1],
                        date=datetime.datetime.now()))
        except Exception as e:
            self.sock.close()
            self.pending_logs.join()
            self.exception_queue.put(e)


class Sender(object):
    def __init__(self, log_transport, pending_logs, exception_queue):
        self.log_transport = log_transport
        self.pending_logs = pending_logs
        self.exception_queue = exception_queue

    def as_thread(self):
        sender = threading.Thread(target=self._send)
        sender.daemon = True
        return sender

    def _send(self):
        try:
            while True:
                self._send_message()
        except Exception as e:
            self.exception_queue.put(e)

    def _send_message(self):
        message = self.pending_logs.get()
        try:
            for retry in retry_loop(15 * 60, delay=1, back_off=1.5):
                try:
                    self.log_transport.send_message(message)
                except elasticsearch.exceptions.ConnectionError as e:
                    retry(e)
        except elasticsearch.exceptions.TransportError as e:
            self._raport_error(message, e)
        self.pending_logs.task_done()

    def _raport_error(self, message, error):
        connection = elasticsearch.Elasticsearch(
            hosts=[configuration['host']],
            use_ssl=configuration['use_ssl'],
            http_auth=configuration['http_auth'])
        body = {'message': str(message), 'error': str(error)}
        connection.index(index='elasticsearch-raven-error', body=body,
                         doc_type='elasticsearch-raven-log')


def retry_loop(timeout, delay, back_off=1.0):
    start_time = time.time()
    exceptions = set()

    def retry(exception):
        exceptions.add(exception)
    yield retry
    while time.time() - start_time <= timeout:
        if not exceptions:
            return
        time.sleep(delay)
        delay *= back_off
        exceptions.clear()
        yield retry

    raise exceptions.pop()



import base64
import collections
import contextlib
import datetime
import hashlib
import itertools
import json
import logging
import re
import sys
import time
import zlib

import elasticsearch

from elasticsearch_raven import configuration
from elasticsearch_raven import exceptions
from elasticsearch_raven.postfix import postfix_encoded_data


class SentryMessage(collections.namedtuple('SentryMessage',
                                           ['headers', 'body'])):
    @classmethod
    def create_from_udp(cls, data):
        try:
            byte_headers, data = data.split(b'\n\n')
        except ValueError:
            raise exceptions.DamagedSentryMessageError
        headers = cls.parse_headers(str(byte_headers.decode('utf-8')))
        data = base64.b64decode(data)
        return cls(headers, data)

    @classmethod
    def create_from_http(cls, raw_headers, data):
        headers = cls.parse_headers(raw_headers)
        data = base64.b64decode(data)
        return cls(headers, data)

    @staticmethod
    def parse_headers(raw_headers):
        match = re.search(r'sentry_key=(?P<sentry_key>[^, =]+), sentry_secret='
                          r'(?P<sentry_secret>[^, =]+)$', raw_headers)
        if match:
            return match.groupdict()
        else:
            raise exceptions.BadSentryMessageHeaderError

    def decode_body(self):
        try:
            return json.loads(zlib.decompress(self.body).decode('utf-8'))
        except (zlib.error, ValueError):
            raise exceptions.DamagedSentryMessageBodyError


def get_configured_log_transport():
    return LogTransport(configuration['host'],
                        use_ssl=configuration['use_ssl'],
                        http_auth=configuration['http_auth'])


class LogTransport:
    DOCUMENT_TYPE = 'raven-log'

    def __init__(self, host, use_ssl=False, http_auth=None):
        self._connection = elasticsearch.Elasticsearch(hosts=[host],
                                                       http_auth=http_auth,
                                                       use_ssl=use_ssl)

    def send_message(self, message):
        message_body = message.decode_body()
        postfix_encoded_data(message_body)
        message_id = hash_dict(message_body)
        index = message_body['project'].format(datetime.datetime.now())
        self.send(message_body, index, message_id)

    def send(self, body, index, message_id):
        for retry in retry_loop(15 * 60, delay=1, back_off=1.5):
            try:
                with logger_level_to_error('elasticsearch'):
                    self._connection.index(body=body, index=index,
                                           id=message_id,
                                           doc_type=self.DOCUMENT_TYPE)
            except elasticsearch.exceptions.ConnectionError as e:
                retry(e)

    def search(self, segment_size=1000, **kwargs):
        for offset in itertools.count(step=segment_size):
            response = self._connection.search(doc_type=self.DOCUMENT_TYPE,
                                               size=segment_size, from_=offset,
                                               **kwargs)
            hits = response['hits']['hits']
            for hit in hits:
                yield hit
            if len(hits) < segment_size:
                break

    def delete(self, index, record_id):
        self._connection.delete(index, self.DOCUMENT_TYPE, record_id)


def hash_dict(dictionary):
    message_json = json.dumps(
        dictionary, indent=None, ensure_ascii=True, separators=(', ', ': '),
        sort_keys=True)
    sha1 = hashlib.sha1()
    sha1.update(message_json.encode('ascii'))
    return sha1.hexdigest()


def retry_loop(timeout, delay, back_off=1.0):
    start_time = time.time()
    exceptions = []

    def retry(exception):
        exceptions.append(exception)

    while True:
        yield retry
        if not exceptions:
            return
        if time.time() - start_time > timeout:
            break
        else:
            exceptions.clear()
        time.sleep(delay)
        delay *= back_off

    raise exceptions[0]


@contextlib.contextmanager
def logger_level_to_error(logger_name):
    logger = logging.getLogger(logger_name)
    level = logger.level
    logger.setLevel(logging.ERROR)
    yield
    logger.setLevel(level)


def update_ids():
    log_transport = get_configured_log_transport()
    all_count, modified_count = 0, 0
    for log in log_transport.search():
        all_count += 1
        log_id = hash_dict(log['_source'])
        if log['_id'] != log_id:
            modified_count += 1
            log_transport.send(log['_source'], log['_index'], log_id)
            log_transport.delete(log['_index'], log['_id'])
    sys.stdout.write('Logs: {}\nModified: {}\n'.format(all_count,
                                                       modified_count))



from setuptools import setup

setup(
    name='Elasticsearch-Raven',
    version='1.2',
    author='Tomasz Wysocki',
    author_email='tomasz@pozytywnie.pl',
    url='https://github.com/pozytywnie/elasticsearch-raven/',
    packages=['elasticsearch_raven'],
    scripts=['bin/elasticsearch-raven.py', 'bin/update_ids.py'],
    license='MIT',
    description='Proxy that allows to send logs from Raven to Elasticsearch.',
    long_description=open('README.rst').read(),
    install_requires=['elasticsearch', 'kombu'],
    test_suite='tests',
)
